//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;

using Smdn.IO.FileSystems;

namespace Smdn.Applications.DirectoryAnalyzer.Analysis {
  public static class DirectoryEntryComparison {
    private interface IComparerImpl {
      int Compare(DirectoryEntry x, DirectoryEntry y);
      int CompareIgnoreType(DirectoryEntry x, DirectoryEntry y);
    }

    private class ComparerBase {
      protected int CompareByType(DirectoryEntry x, DirectoryEntry y)
      {
        if (x is Directory && y is File) return 1;
        if (x is File && y is Directory) return -1;

        return 0;
      }
    }

    private class NameComparer : ComparerBase, IComparerImpl {
      public static readonly NameComparer Ascending = new NameComparer(+1);
      public static readonly NameComparer Descending = new NameComparer(-1);

      private readonly int order;

      private NameComparer(int order)
      {
        this.order = order;
      }

      public int Compare(DirectoryEntry x, DirectoryEntry y)
      {
        var comp = CompareByType(x, y);

        if (comp == 0)
          return order * String.Compare(x.Name, y.Name/*, StringComparison.Ordinal*/); // XXX
        else
          return order * comp;
      }

      public int CompareIgnoreType(DirectoryEntry x, DirectoryEntry y)
      {
        return order * String.Compare(x.Name, y.Name/*, StringComparison.Ordinal*/); // XXX
      }
    }

    private class FullNameComparer : ComparerBase, IComparerImpl {
      public static readonly FullNameComparer Ascending = new FullNameComparer(+1);
      public static readonly FullNameComparer Descending = new FullNameComparer(-1);

      private readonly int order;

      private FullNameComparer(int order)
      {
        this.order = order;
      }

      public int Compare(DirectoryEntry x, DirectoryEntry y)
      {
        var comp = CompareByType(x, y);

        if (comp == 0)
          return order * String.Compare(x.FullName, y.FullName/*, StringComparison.Ordinal*/); // XXX
        else
          return order * comp;
      }

      public int CompareIgnoreType(DirectoryEntry x, DirectoryEntry y)
      {
        return order * String.Compare(x.FullName, y.FullName/*, StringComparison.Ordinal*/); // XXX
      }
    }

    private class SizeComparer : ComparerBase, IComparerImpl {
      public static readonly SizeComparer Ascending = new SizeComparer(+1);
      public static readonly SizeComparer Descending = new SizeComparer(-1);

      private readonly int order;

      private SizeComparer(int order)
      {
        this.order = order;
      }

      public int Compare(DirectoryEntry x, DirectoryEntry y)
      {
        var comp = CompareByType(x, y);

        if (comp == 0)
          return CompareIgnoreType(x, y);
        else
          return order * comp;
      }

      public int CompareIgnoreType(DirectoryEntry x, DirectoryEntry y)
      {
        long sizeX, sizeY;

        var fx = x as File;
        var fy = y as File;

        if (fx == null)
          sizeX = (x as DirectoryTree).GetTotalFileSize(); // can't compare Directory!
        else
          sizeX = fx.Length;

        if (fy == null)
          sizeY = (y as DirectoryTree).GetTotalFileSize(); // can't compare Directory!
        else
          sizeY = fy.Length;

        if (sizeX == sizeY)
          // サイズが同じ場合は、名前で比較する
          return order * String.Compare(x.Name, y.Name/*, StringComparison.Ordinal*/); // XXX
        else
          return (sizeX < sizeY) ? order : -order; // inverted; larger one first
      }
    }

    public static Comparison<DirectoryEntry> GetComparison(DirectoryEntryCompareMode compareMode)
    {
      return GetComparison(compareMode, DirectoryEntryCompareOrder.Ascending, true);
    }

    public static Comparison<DirectoryEntry> GetComparison(DirectoryEntryCompareMode compareMode,
                                                           DirectoryEntryCompareOrder compareOrder,
                                                           bool compareEntryType)
    {
      IComparerImpl comparer;

      if (compareOrder == DirectoryEntryCompareOrder.Ascending) {
        switch (compareMode) {
          case DirectoryEntryCompareMode.BySize:      comparer = SizeComparer.Ascending; break;
          case DirectoryEntryCompareMode.ByName:      comparer = NameComparer.Ascending; break;
          case DirectoryEntryCompareMode.ByFullName:  comparer = FullNameComparer.Ascending; break;
          default: return null;
        }
      }
      else if (compareOrder == DirectoryEntryCompareOrder.Descending) {
        switch (compareMode) {
          case DirectoryEntryCompareMode.BySize:      comparer = SizeComparer.Descending; break;
          case DirectoryEntryCompareMode.ByName:      comparer = NameComparer.Descending; break;
          case DirectoryEntryCompareMode.ByFullName:  comparer = FullNameComparer.Descending; break;
          default: return null;
        }
      }
      else {
        return null;
      }

      if (compareEntryType)
        return comparer.Compare;
      else
        return comparer.CompareIgnoreType;
    }
  }
}
