//  
//  Author:
//       smdn <smdn@smdn.jp>
// 
//  Copyright (c) 2003-2013 smdn
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
// 
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.IO;

namespace Smdn.Applications.DirectoryAnalyzer.Rendering {
  using Graphing;

  public class RenderingContext : TaskContext, ITaskContext
  {
    public RenderingContext()
    {
      this.parameter = new RenderingParameter();
      this.status = new RenderingStatus();
      this.result = new RenderingResult();
    }

    #region parameters
    private RenderingParameter parameter = null;

    public RenderingParameter Parameter
    {
      get
      {
        return this.parameter;
      }
    }

    public static class RenderingParameterDefaults {
      public static readonly int OutputWidth = 800;
      public static readonly int OutputHeight = 800;
      public static readonly float InnerDiameterPercent = 0.35f;
      public static readonly ColoringMode ColoringMode = ColoringMode.Directory;
      public static readonly int ColoringHue = 15;
      public static readonly bool DrawNames = true;
      public static readonly bool DrawStatistics = true;
      public static readonly bool DrawSmallEntries = true;
      public static readonly int FontSize = 8;
      public static readonly SmoothingMode SmoothingMode = SmoothingMode.AntiAlias;
      public static readonly TextRenderingHint TextRenderingHint = TextRenderingHint.AntiAlias;
      public static readonly InterpolationMode InterpolationMode = InterpolationMode.HighQualityBicubic;
      public static readonly Color BackgroundColor = Color.White;
      public static readonly Color TextColor = Color.Black;
    }

    public class RenderingParameter : TaskParameter
    {
      public Graph Graph = null;
      public int MaxDepth = 0;
      public int OutputWidth                      = RenderingParameterDefaults.OutputWidth;
      public int OutputHeight                     = RenderingParameterDefaults.OutputHeight;
      public float InnerDiameterPercent           = RenderingParameterDefaults.InnerDiameterPercent;
      public ColoringMode ColoringMode            = RenderingParameterDefaults.ColoringMode;
      public int ColoringHue                      = RenderingParameterDefaults.ColoringHue;
      public bool DrawNames                       = RenderingParameterDefaults.DrawNames;
      public bool DrawStatistics                  = RenderingParameterDefaults.DrawStatistics;
      public bool DrawSmallEntries                = RenderingParameterDefaults.DrawSmallEntries;
      public int FontSize                         = RenderingParameterDefaults.FontSize;
      public SmoothingMode SmoothingMode          = RenderingParameterDefaults.SmoothingMode;
      public TextRenderingHint TextRenderingHint  = RenderingParameterDefaults.TextRenderingHint;
      public InterpolationMode InterpolationMode  = RenderingParameterDefaults.InterpolationMode;
      public Color BackgroundColor                = RenderingParameterDefaults.BackgroundColor;
      public Color TextColor                      = RenderingParameterDefaults.TextColor;
    }
    #endregion

    #region status
    private RenderingStatus status = null;

    public RenderingStatus Status
    {
      get
      {
        return this.status;
      }
    }

    public class RenderingStatus : TaskStatus
    {
      public uint TotalEntryCount = 0;
      public uint RenderedEntryCount = 0;

      public override uint Total
      {
        get
        {
          return this.TotalEntryCount;
        }
      }

      public override uint Processed
      {
        get
        {
          return this.RenderedEntryCount;
        }
      }
    }
    #endregion

    #region results
    private RenderingResult result = null;

    public RenderingResult Result
    {
      get
      {
        return this.result;
      }
    }

    public class RenderingResult : TaskResult
    {
      public Bitmap GraphImage = null;

      public void Save(string path)
      {
        using (var stream = new FileStream(path, FileMode.Create, FileAccess.Write, FileShare.None)) {
          Save(stream);
        }
      }

      public void Save(FileStream stream)
      {
        GraphImage.Save(stream, GetImageFormat(stream));
      }

      private ImageFormat GetImageFormat(FileStream stream)
      {
        switch (Path.GetExtension(stream.Name).ToLowerInvariant()) {
          case ".jpg":
          case ".jpeg":
            return ImageFormat.Jpeg;

          case ".bmp":
            return ImageFormat.Bmp;

          case ".png":
          default:
            return ImageFormat.Png;
        }
      }
    }
    #endregion

    #region ITaskContext

    TaskContext.TaskParameter ITaskContext.Parameter
    {
      get { return this.parameter; }
    }

    TaskContext.TaskResult ITaskContext.Result
    {
      get { return this.result; }
    }

    TaskContext.TaskStatus ITaskContext.Status
    {
      get { return this.status; }
    }

    #endregion
  }
}
